/* $Id: rb_mbus.c,v 1.3 2001/12/17 03:18:41 kurosawa Exp $ */

/*
 * Copyright (C) 2001 FUJITSU LABORATRIES LTD.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "rb_mbus.h"

/* #define RBMBUS_DEBUG */
#ifdef RBMBUS_DEBUG
#define DPRINT(ARG)  rbmbus_dprint ARG
static void rbmbus_dprint(char *fmt, ...);
#else
#define DPRINT(ARG)
#endif

static VALUE cMbus;

static ID id_mbusdata;
static ID id_cmd_handler;
static ID id_err_handler;
static ID id_call;

static rbmbus_t *
rbmbus_get_data(mbusobj)
	VALUE mbusobj;
{
	rbmbus_t *rmbus;

	Data_Get_Struct(mbusobj, rbmbus_t, rmbus);
	return rmbus;
}

static void
rbmbus_cmd_handler(src, cmd, arg, dat)
	char *src, *cmd, *arg;
	void *dat;
{
	rbmbus_hdl_data_t *hdata;
	VALUE rsrc, rcmd, rarg, f;

	DPRINT(("mbus.cmd_handler src=%s dat=%p %s %s\n", src, dat, cmd, arg));
	hdata = dat;
	f = rb_ivar_get(hdata->mbdata->obj, id_cmd_handler);
	if (TYPE(f) == T_NIL)
		return;

	rsrc = rb_str_new2(src);
	rcmd = rb_str_new2(cmd);
	rarg = rb_str_new2(arg);
	rb_funcall(f, id_call, 4, rsrc, rcmd, rarg, hdata->udata);
}

static void
rbmbus_err_handler(seq, reason)
	int seq, reason;
{
	/* XXX not implemented. Can't get the Mbus instance. */
}

static VALUE
rbmbus_new(self, cmd_hdl, err_hdl, src)
	VALUE self, cmd_hdl, err_hdl, src;
{
	VALUE data;
	rbmbus_t *rmbus;

	Check_Type(src, T_STRING);

	rmbus = xmalloc(sizeof(rbmbus_t));
	memset(rmbus, 0, sizeof(*rmbus));
	rmbus->mb = mbus_init(rbmbus_cmd_handler, rbmbus_err_handler,
			      STR2CSTR(src));

	/* XXX  Should add mark & free functions */
	data = Data_Wrap_Struct(cMbus, 0, 0, rmbus);
	rb_ivar_set(data, id_cmd_handler, cmd_hdl);
	rb_ivar_set(data, id_err_handler, err_hdl);
	rmbus->obj = data;

	return data;
}

static VALUE
rbmbus_heartbeat(self, interval)
	VALUE self, interval;
{
	rbmbus_t *rmbus;

	Check_Type(interval, T_FIXNUM);
	rmbus = rbmbus_get_data(self);
	mbus_heartbeat(rmbus->mb, NUM2INT(interval));

	return Qnil;
}

static VALUE
rbmbus_recv(self, data, timeout)
	VALUE self, data, timeout;
{
	VALUE rbret;
	rbmbus_t *rmbus;
	rbmbus_hdl_data_t hdata;
	struct timeval tv;
	int ret;
	int t;

	Check_Type(timeout, T_FIXNUM);
	rmbus = rbmbus_get_data(self);
	hdata.mbdata = rmbus;
	hdata.udata = data;

	t = NUM2INT(timeout);
	tv.tv_sec = t / 1000;
	tv.tv_usec = (t % 1000) * 1000;

	ret = mbus_recv(rmbus->mb, &hdata, &tv);
	if (ret)
		rbret = Qtrue;
	else
		rbret = Qfalse;

	return rbret;
}

static VALUE
rbmbus_qmsg(self, dest, cmd, args, isreliable)
	VALUE self, dest, cmd, args, isreliable;
{
	rbmbus_t *rmbus;
	int r;

	Check_Type(dest, T_STRING);
	Check_Type(cmd, T_STRING);
	Check_Type(args, T_STRING);
	Check_Type(isreliable, T_TRUE | T_FALSE);

	rmbus = rbmbus_get_data(self);
	if (TYPE(isreliable) == T_TRUE)
		r = 1;
	else
		r = 0;

	mbus_qmsg(rmbus->mb, STR2CSTR(dest), STR2CSTR(cmd), STR2CSTR(args),
		  r);

	return Qnil;
}

static VALUE
rbmbus_send(self)
	VALUE self;
{
	rbmbus_t *rmbus;

	rmbus = rbmbus_get_data(self);
	mbus_send(rmbus->mb);
	return Qnil;
}

static VALUE
rbmbus_retransmit(self)
	VALUE self;
{
	rbmbus_t *rmbus;

	rmbus = rbmbus_get_data(self);
	mbus_retransmit(rmbus->mb);
	return Qnil;
}

void
Init_mbus()
{
	cMbus = rb_define_class("Mbus", rb_cData);
	id_mbusdata = rb_intern("mbusdata");
	id_cmd_handler = rb_intern("cmd_handler");
	id_err_handler = rb_intern("err_handler");
	id_call = rb_intern("call");
	rb_define_singleton_method(cMbus, "new", rbmbus_new, 3);

	/* Instance methods */
	rb_define_method(cMbus, "heartbeat", rbmbus_heartbeat, 1);
	rb_define_method(cMbus, "recv", rbmbus_recv, 2);
	rb_define_method(cMbus, "qmsg", rbmbus_qmsg, 4);
	rb_define_method(cMbus, "send", rbmbus_send, 0);
	rb_define_method(cMbus, "retransmit", rbmbus_retransmit, 0);
}


#ifdef RBMBUS_DEBUG
#include <stdarg.h>

static void
rbmbus_dprint(char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}
#endif
