#! /usr/local/bin/ruby -w
# $Id: xcgroupsrv.cgi,v 1.2 2001/10/09 00:43:15 kurosawa Exp $

# Copyright (C) 2001 FUJITSU LABORATRIES LTD.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the name of the project nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.

require 'cgi'

XCGroupFileDir = '/usr/local/share/xcgroup'
XCGroupTimeout = 180		# 3 minutes

# Group file format (text/plain version):
# <ipv6-address> wsp+ <timestamp> wsp+ options...
# ...
# 

class GroupFile
  def initialize(file)
    @members = []
    @f = open(file, "a+")
    begin
      @f.flock(File::LOCK_EX)
      @f.seek(IO::SEEK_SET, 0)
      parse
    rescue
      $stderr.print("GroupFile Error: #$!\n")
      @f.flock(File::LOCK_UN)
      @f.close
    end
  end

  def parse
    t = Time::now.to_i
    @f.each do |l|
      l.sub!('#.*$', '')
      l.chomp!
      m = l.split
      if ((m != []) && (m[1].to_i + XCGroupTimeout > t))
	@members.push(m)
      end
    end
  end

  def lookup(addr)
    @members.assoc(addr)
  end

  def add_addr(addr)
    t = Time::now.to_i.to_s
    m = lookup(addr)
    if (m.nil?)
      @members.push([addr, t])
    else
      @members[@members.index(m)][1] = t
    end
    @members
  end

  def remove_addr(addr)
    m = lookup(addr)
    if (m)
      @members.delete(m)
    end
  end

  def write
    @f.truncate(0)
    @f.printf("# Last-Modified: %s\n", Time::now)
    @members.each do |m|
      @f.printf("%s\n", m.join("\t"))
    end
  end

  def show
    @members.each do |m|
      printf("%s %s\n", m[0], m[1])
    end
  end

  def each_addr
    @members.each do |m|
      yield(m[0])
    end
  end

  def close
    write
    @f.flock(File::LOCK_UN)
    @f.close
  end
end


#### main

cgi = CGI.new

$stdout.print("Content-Type: text/plain\n\n")

if cgi.has_key?('cmd')
  cmd = cgi['cmd'][0].downcase
else
  cmd = 'join'
end

if cgi.has_key?('group')
  groupname = cgi['group'][0]
else
  $stdout.print("!ERROR\nURL specification error!\n")
  exit(1)
end


if cgi.has_key?('addr')
  addr = cgi['addr'][0]
else
  addr = ENV['REMOTE_ADDR']
end

if addr =~ /^[0-9.]+$/
  $stdout.print("!ERROR\nConnection from IPv4 address.\n")
  exit(1)
end

group = GroupFile.new(XCGroupFileDir + '/' + groupname + '.txt')
begin
  case cmd
  when 'join'
    group.add_addr(addr)
  when 'leave'
    group.remove_addr(addr)
  end
  
  group.each_addr do |a|
    $stdout.printf("%s\n", a)
  end
ensure
  group.close
end
